/*
THE COMPUTER CODE CONTAINED HEREIN IS THE SOLE PROPERTY OF PARALLAX
SOFTWARE CORPORATION ("PARALLAX").  PARALLAX, IN DISTRIBUTING THE CODE TO
END-USERS, AND SUBJECT TO ALL OF THE TERMS AND CONDITIONS HEREIN, GRANTS A
ROYALTY-FREE, PERPETUAL LICENSE TO SUCH END-USERS FOR USE BY SUCH END-USERS
IN USING, DISPLAYING,  AND CREATING DERIVATIVE WORKS THEREOF, SO LONG AS
SUCH USE, DISPLAY OR CREATION IS FOR NON-COMMERCIAL, ROYALTY OR REVENUE
FREE PURPOSES.  IN NO EVENT SHALL THE END-USER USE THE COMPUTER CODE
CONTAINED HEREIN FOR REVENUE-BEARING PURPOSES.  THE END-USER UNDERSTANDS
AND AGREES TO THE TERMS HEREIN AND ACCEPTS THE SAME BY USE OF THIS FILE.  
COPYRIGHT 1993-1998 PARALLAX SOFTWARE CORPORATION.  ALL RIGHTS RESERVED.
*/
/*
 * $Source: BigRed:miner:source:bios::RCS:mouse.c $
 * $Revision: 1.8 $
 * $Author: allender $
 * $Date: 1996/02/21 13:57:36 $
 * 
 * Functions to access Mouse and Cyberman...
 * 
 * $Log: mouse.c $
 * Revision 1.8  1996/02/21  13:57:36  allender
 * cursor device manager stuff added here so as not to
 * rely on InterfaceLib anymore
 *
 * Revision 1.7  1995/10/17  15:42:21  allender
 * new mouse function to determine single button press
 *
 * Revision 1.6  1995/10/03  11:27:31  allender
 * fixed up hotspot problems with the mouse on multiple monitors
 *
 * Revision 1.5  1995/07/13  11:27:08  allender
 * trap button checks at MAX_MOUSE_BUTTONS
 *
 * Revision 1.4  1995/06/25  21:56:53  allender
 * added events include
 *
 * Revision 1.3  1995/05/11  17:06:38  allender
 * fixed up mouse routines
 *
 * Revision 1.2  1995/05/11  13:05:53  allender
 * of mouse handler code
 *
 * Revision 1.1  1995/05/05  09:54:45  allender
 * Initial revision
 *
 * Revision 1.9  1995/01/14  19:19:52  john
 * Fixed signed short error cmp with -1 that caused mouse
 * to break under Watcom 10.0
 * 
 * Revision 1.8  1994/12/27  12:38:23  john
 * Made mouse use temporary dos buffer instead of
 * 
 * allocating its own.
 * 
 * 
 * Revision 1.7  1994/12/05  23:54:53  john
 * Fixed bug with mouse_get_delta only returning positive numbers..
 * 
 * Revision 1.6  1994/11/18  23:18:18  john
 * Changed some shorts to ints.
 * 
 * Revision 1.5  1994/09/13  12:34:02  john
 * Added functions to get down count and state.
 * 
 * Revision 1.4  1994/08/29  20:52:19  john
 * Added better cyberman support; also, joystick calibration
 * value return funcctiionn,
 * 
 * Revision 1.3  1994/08/24  18:54:32  john
 * *** empty log message ***
 * 
 * Revision 1.2  1994/08/24  18:53:46  john
 * Made Cyberman read like normal mouse; added dpmi module; moved
 * mouse from assembly to c. Made mouse buttons return time_down.
 * 
 * Revision 1.1  1994/08/24  13:56:37  john
 * Initial revision
 * 
 * 
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <ggi/libggi.h>

#include "error.h"
#include "fix.h"
#include "mouse.h"
#include "gtimer.h"
#include "macsys.h"

#include "dtypes.h"

#include "debug.h"

#define ME_CURSOR_MOVED	(1<<0)
#define ME_LB_P 			(1<<1)
#define ME_LB_R 			(1<<2)
#define ME_RB_P 			(1<<3)
#define ME_RB_R 			(1<<4)
#define ME_MB_P 			(1<<5)
#define ME_MB_R 			(1<<6)
#define ME_OB_P 			(1<<7)
#define ME_OB_R 			(1<<8)
#define ME_X_C 			(1<<9)
#define ME_Y_C 			(1<<10)
#define ME_Z_C 			(1<<11)
#define ME_P_C 			(1<<12)
#define ME_B_C 			(1<<13)
#define ME_H_C 			(1<<14)
#define ME_O_C 			(1<<15)

#define MOUSE_MAX_BUTTONS	11

typedef struct event_info {
	short x;
	short y;
	short z;
	short pitch;
	short bank;
	short heading;
	ushort button_status;
	ushort device_dependant;
} event_info;

typedef struct mouse_info {
	fix		ctime;
	ubyte	cyberman;
	int		num_buttons;
	ubyte	pressed[MOUSE_MAX_BUTTONS];
	fix		time_went_down[MOUSE_MAX_BUTTONS];
	fix		time_held_down[MOUSE_MAX_BUTTONS];
	uint	num_downs[MOUSE_MAX_BUTTONS];
	uint	num_ups[MOUSE_MAX_BUTTONS];
//	ubyte	went_down; /* Not in PC version, not needed with 'num_downs' etc */
	event_info *x_info;
	ushort	button_status;
} mouse_info;

typedef struct cyberman_info {
	ubyte device_type;
	ubyte major_version;
	ubyte minor_version;
	ubyte x_descriptor;
	ubyte y_descriptor;
	ubyte z_descriptor;
	ubyte pitch_descriptor;
	ubyte roll_descriptor;
	ubyte yaw_descriptor;
	ubyte reserved;
} cyberman_info;

static mouse_info Mouse;

static int Mouse_installed = 0;


//GGI data:
extern  ggi_visual_t		visual;
extern  ggi_directbuffer_t		dbuf;	// GGI direct acces to screen memory
extern  ggi_pixellinearbuffer	*plb;

//Mouse globals
static double mouse_x, mouse_y;
static double mouse_saved_x, mouse_saved_y; //used when hiding/unhiding to reset the real (displayed) postion
double mouse_accel=1.0;
bool   showmouse=False;

void DrawMouse(void);
void EraseMouse(void);
void MoveMouse(/*int button,*/ int x, int y);

#define WIN_WIDTH 640
#define WIN_HEIGHT 480
#define SCR_WIDTH 640
#define SCR_HEIGHT 480



void mouse_handler()
{
#if 0
	int button_state;

	Mouse.ctime = timer_get_fixed_seconds();
	button_state = Button();
	
	if (button_state) {
		if (!Mouse.pressed[MB_LEFT]) {
			Mouse.pressed[MB_LEFT] = 1;
			Mouse.time_went_down[MB_LEFT] = Mouse.ctime;
			Mouse.went_down = 1;
		}
		Mouse.num_downs[MB_LEFT]++;
	} else {
		if (Mouse.pressed[MB_LEFT]) {
			Mouse.pressed[MB_LEFT] = 0;
			Mouse.time_held_down[MB_LEFT] += Mouse.ctime-Mouse.time_went_down[MB_LEFT];
			Mouse.num_ups[MB_LEFT]++;
			Mouse.went_down = 0;
		}
	}
#else
	ggi_event event;
	ggi_event_mask event_mask;
	struct timeval tv;
	int mouse_dx, mouse_dy;

	Mouse.ctime = timer_get_fixed_seconds();

	event_mask= (ggi_event_mask) (/*emKeyPress | emKeyRelease |*/ emPtrRelative | emPtrButtonPress | emPtrButtonRelease);
    tv.tv_sec = 0;
    tv.tv_usec = 0;

	while (ggiEventPoll(visual, event_mask, &tv)) {
		ggiEventRead(visual, &event, event_mask);
		switch(event.any.type) {
		case evPtrRelative:	/* pointer movements reported relative	*/
			DEB1(prt("GGI event: pointer rel motion dx=%d, dy=%d\n", event.pmove.x, event.pmove.y));
			mouse_dx=event.pmove.x; //* factor ? */
			mouse_dy=event.pmove.y; //* factor ? */
//move_mouse(mouse_dx, mouse_dy);
DEB1(fprintf(stderr, "GGI mouse_handler, dx,dy=%d,%d\n", mouse_dx, mouse_dy));

DEB1(fprintf(stderr, "GGI mouse_handler (1), x,y=%d,%d\n", (int) mouse_x, (int) mouse_y));
			MoveMouse(mouse_dx, mouse_dy);
DEB1(fprintf(stderr, "GGI mouse_handler (2), x,y=%d,%d\n", (int) mouse_x, (int) mouse_y));
			break;
				//	evPtrAbsolute:	/* pointer movements reported absolute	*/
				//	    break;
		case evPtrButtonPress:	/* pointer button pressed	*/
			DEB1(prt("GGI event: button press (%d)\n", event.pbutton.button));
			{int button=event.pbutton.button;
			if (button & 1) { // left button pressed
				if (!Mouse.pressed[MB_LEFT])	{
					Mouse.pressed[MB_LEFT] = 1;
					Mouse.time_went_down[MB_LEFT] = Mouse.ctime;
//			Mouse.went_down = 1; /* only in mac version .. */
				}
			Mouse.num_downs[MB_LEFT]++;
			}
			if (button & 4) { // middle button pressed
				if (!Mouse.pressed[MB_MIDDLE])	{
					Mouse.pressed[MB_MIDDLE] = 1;
					Mouse.time_went_down[MB_MIDDLE] = Mouse.ctime;
				}
			Mouse.num_downs[MB_MIDDLE]++;
			}
			if (button & 2) { // right button pressed
				if (!Mouse.pressed[MB_RIGHT])	{
					Mouse.pressed[MB_RIGHT] = 1;
					Mouse.time_went_down[MB_RIGHT] = Mouse.ctime;
				}
			Mouse.num_downs[MB_RIGHT]++;
			}
			}
			break;
		case evPtrButtonRelease:
			DEB1(prt("GGI event: button release (%d)\n", event.pbutton.button));
			{int button=event.pbutton.button;
			if (button & 1) {
				if (Mouse.pressed[MB_LEFT])	{
					Mouse.pressed[MB_LEFT] = 0;
					Mouse.time_held_down[MB_LEFT] += Mouse.ctime-Mouse.time_went_down[MB_LEFT];
				}
				Mouse.num_ups[MB_LEFT]++;
			}
			if (button & 4) {
				if (Mouse.pressed[MB_MIDDLE])	{
					Mouse.pressed[MB_MIDDLE] = 0;
					Mouse.time_held_down[MB_MIDDLE] += Mouse.ctime-Mouse.time_went_down[MB_MIDDLE];
				}
				Mouse.num_ups[MB_MIDDLE]++;
			}
			if (button & 2) {
				if (Mouse.pressed[MB_RIGHT])	{
					Mouse.pressed[MB_RIGHT] = 0;
					Mouse.time_held_down[MB_RIGHT] += Mouse.ctime-Mouse.time_went_down[MB_RIGHT];
				}
				Mouse.num_ups[MB_RIGHT]++;
			}
			}
			break;
		default: continue;
		}
	}
#endif
}

void mouse_flush()
{
	int i;
	fix CurTime;
	
	if (!Mouse_installed)
		return;

//	_disable();
	CurTime = timer_get_fixed_seconds();
	for (i = 0; i < MOUSE_MAX_BUTTONS; i++) {
		Mouse.pressed[i] = 0;
		Mouse.time_went_down[i] = CurTime;
		Mouse.time_held_down[i] = 0;
		Mouse.num_downs[i] = 0;
		Mouse.num_ups[i] = 0;
	}
//	Mouse.went_down = 0; /* mac only */
//	_enable();
}

#if 0 /* Not used */
void mouse_set_center()
{
//	center_p.h =200; //should be .5*window width (WHS)
//	center_p.v =100; //should be .5*window height
	mouse_x=(double) (SCR_WIDTH/2);
	mouse_x=(double) (SCR_HEIGHT/2);
}
#endif

void mouse_close()
{
	if (Mouse_installed)
		Mouse_installed = 0;
}



int mouse_init()
{
	if (Mouse_installed)
		return Mouse.num_buttons;

	//WHS: How do I get the number of buttons ?
	Mouse.num_buttons = 3;

	Mouse_installed = 1;
	atexit(mouse_close);
	mouse_flush();
//	mouse_set_center();
	
	return Mouse.num_buttons;
}

#if 0 /* Not used */
void mouse_set_limits( int x1, int y1, int x2, int y2 )
{
	mouse_hold_left=x1;
	mouse_hold_right=x2;
	mouse_hold_top=y1; /* -mouse_width ? */
	mouse_hold_bottom=y2; /* -mouse_height ? */
}
#endif
//WHS: added this, not needed any more I think.
void mouse_center() {
if (showmouse) EraseMouse();
mouse_x=mouse_saved_x=WIN_WIDTH/2;
mouse_y=mouse_saved_y=WIN_HEIGHT/2;
if (showmouse) DrawMouse();
}

void mouse_get_pos( int *x, int *y)
{

mouse_handler(); //temp

	*x=(int) mouse_x;
	*y=(int) mouse_y;
}

void mouse_get_delta( int *dx, int *dy )
{
//printf("mouse_get_delta\n");
	if (!Mouse_installed) {
		*dx = *dy = 0;
		return;
	}
mouse_handler(); //temp

	*dx = (int) mouse_x - WIN_WIDTH/2;
	*dy = (int) mouse_y - WIN_HEIGHT/2;

	//Now reset the mouse position to the center of the screen
	if (showmouse) EraseMouse();
	mouse_x = (double) WIN_WIDTH/2;
	mouse_y = (double) WIN_HEIGHT/2;
	if (showmouse) DrawMouse();
}

void mouse_get_delta_no_reset( int *dx, int *dy )
{	
	if (!Mouse_installed) {
		*dx = *dy = 0;
		return;
	}
mouse_handler(); //temp

	*dx = (int) mouse_x - WIN_WIDTH/2;
	*dy = (int) mouse_y - WIN_HEIGHT/2;
}

int mouse_get_btns()
{
	int  i;
	uint flag=1;
	int  status = 0;

	if (!Mouse_installed)
		return 0;

mouse_handler(); //temp

	for (i = 0; i < MOUSE_MAX_BUTTONS; i++) {
		if (Mouse.pressed[i])
			status |= flag;
		flag <<= 1;
	}
	return status;
}

#if 0
void mouse_set_pos( int x, int y)
{

}
#endif
/* This should be replaced with mouse_button_down_count(int button)	*/
int mouse_went_down(int button)
{
	int count;

	if (!Mouse_installed)
		return 0;

mouse_handler(); //temp

	if ((button < 0) || (button >= MOUSE_MAX_BUTTONS))
		return 0;

//	_disable();		
#if 0
	count = (int)Mouse.went_down;
	Mouse.went_down = 0;
#else
	count = Mouse.num_downs[button];
	Mouse.num_downs[button] = 0;
#endif

// 	_enable();
	return count;
}

// Returns how many times this button has went down since last call.
int mouse_button_down_count(int button)	
{
	int count;

	if (!Mouse_installed)
		return 0;

mouse_handler(); //temp

	if ((button < 0) || (button >= MOUSE_MAX_BUTTONS))
		return 0;

//	_disable();
	count = Mouse.num_downs[button];
	Mouse.num_downs[button] = 0;
//	_enable();
	return count;
}

// Returns 1 if this button is currently down
int mouse_button_state(int button)
{
	int state;

	if (!Mouse_installed)
		return 0;

mouse_handler(); //temp

	if ((button < 0) || (button >= MOUSE_MAX_BUTTONS))
		return 0;

//	_disable();
	state = Mouse.pressed[button];
//	_enable();
	return state;
}

// Returns how long this button has been down since last call.
fix mouse_button_down_time(int button)	
{
	fix time_down, time;

	if (!Mouse_installed)
		return 0;

mouse_handler(); //temp

	if ((button < 0) || (button >= MOUSE_MAX_BUTTONS))
		return 0;

//	_disable();
	if (!Mouse.pressed[button]) {
		time_down = Mouse.time_held_down[button];
		Mouse.time_held_down[button] = 0;
	} else {
		time = timer_get_fixed_seconds();
		time_down = time - Mouse.time_held_down[button];
		Mouse.time_held_down[button] = 0;
	}
//	_enable();

	return time_down;
}

void mouse_get_cyberman_pos( int *x, int *y )
{
}



//WHS: I made this :)
void hide_cursor()
{
DEB1(fprintf(stderr, "hide cursor\n"));
    //WHS: implement with bitmap for GGI/DGA, mouse for X11
	if (showmouse) { mouse_saved_x=mouse_x; mouse_saved_y=mouse_y; EraseMouse(); }
//if (showmouse) printf("hide cursor: x,y=%f,%f; saved_x=%f, saved_y=%f\n", mouse_x, mouse_y, mouse_saved_x, mouse_saved_y);
	showmouse=False;
}

void show_cursor()
{
    //WHS: implement with bitmap for GGI/DGA, mouse for X11
DEB1(fprintf(stderr, "show cursor\n"));
//if (!showmouse) printf("show cursor: x,y=%f,%f; saved_x=%f, saved_y=%f\n", mouse_x, mouse_y, mouse_saved_x, mouse_saved_y);
	if (!showmouse) { mouse_x=mouse_saved_x; mouse_y=mouse_saved_y; DrawMouse(); }
	showmouse=True;
}


/* New mouse pointer stuff for GGI/DGA */
#include "cursor.h" /* cursor and mask */

typedef struct Sprite {
	ushort width;
	ushort height;
	byte *pixels;
	byte *mask;
} Sprite;

Sprite mouse_sprite = { cursor_width, cursor_height, cursor_bits, cursor_mask_bits};

//byte *behind_mouse;
byte behind_mouse[cursor_width*cursor_height];


void DrawMouse(void) {
	byte *m_pixels = mouse_sprite.pixels;
	byte *m_mask = mouse_sprite.mask;
	int  m_width = mouse_sprite.width;
	int  m_height = mouse_sprite.height;
	int  m_x=(int)mouse_x, m_y=(int)mouse_y;
	int  row, col, offset;
	byte new_area[m_width*m_height];
	ggi_color m_colour;

	byte *display_mem=(byte *) plb->read;
	int  black_white[2];
//printf("DrawMouse\n");
	/* Save what's behind the mouse */
	for ( row=0; row<m_height; row++ ) {
		memcpy(&behind_mouse[row*m_width], 
			   &display_mem[(m_y+row)*plb->stride+m_x], m_width);
	}
	memcpy(new_area, behind_mouse, m_width*m_height);

	/* map black and white */
	m_colour.r=0; m_colour.g=0; m_colour.b=0; //black
	black_white[1]=ggiMapColor(visual, &m_colour);
	m_colour.r=65535; m_colour.g=65535; m_colour.b=65535; //white
	black_white[0]=ggiMapColor(visual, &m_colour);

	/* Draw the mouse in the new area */
	for ( row=0; row<m_height; ++row ) {
		for ( col=0; col<m_width; col++ ) {
			offset = row*((m_width+7)>>3)+(col>>3);
            if ( (m_mask[offset]>>(col & 7)) & 0x01 ) {
				new_area[row*m_width+col] = black_white[(m_pixels[offset]>>(col & 7)) & 0x01];
			}
		}
	}
	ggiPutBox(visual, m_x, m_y, m_width, m_height, new_area);
}


void EraseMouse(void) {
//printf("EraseMouse\n");
	ggiPutBox(visual, (int)mouse_x, (int)mouse_y, mouse_sprite.width,
			  mouse_sprite.height, behind_mouse);
}

//Should add a mode, relative, absolute
#define MOVE_REL 0
#define MOVE_ABS 1
//void MoveMouse(int mode, int x, int y) {

void MoveMouse(int x, int y) {
//fprintf(stderr, "MoveMouse by (%d, %d)\n", x,y);
	if ( showmouse )
		EraseMouse();
	mouse_x += (mouse_accel*x);
	if ( mouse_x < 0 )
		mouse_x = 0;
	if ( mouse_x > (SCR_WIDTH-mouse_sprite.width) )
		mouse_x = (SCR_WIDTH-mouse_sprite.width);
	mouse_y += (mouse_accel*y);
	if ( mouse_y < 0 )
		mouse_y = 0;
	if ( mouse_y > (SCR_HEIGHT-mouse_sprite.height) )
		mouse_y = (SCR_HEIGHT-mouse_sprite.height);

	if ( showmouse )
		DrawMouse();
}
